@@ -1,4 +1,3 @@
-#!/opt/advent/bin/python2.7
 # coding: utf-8
 """
 Xacto CLI
@@ -13,13 +12,14 @@ from __future__ import print_function
 
 import sys, os
 import warnings
+import operator
+import collections
 from os import path as pth
 
 import argparse, pkgutil, types, inspect, imp, re
-from itertools import imap, izip, ifilter, repeat, chain
 
 
-__version__ = '0.6.3'
+__version__ = '0.8.2'
 
 
 nil = type('nil', (int,), dict())(0)
@@ -178,9 +178,6 @@ class feature(mapo):
         if keys is NotImplemented:
             return keys
 
-        get = lambda x: None
-        if hasattr(other, 'keys'):
-            get = other.__getitem__
         for key in keys - ikeys:
             self[key] = get(key)
         for key in ikeys - keys:
@@ -193,6 +190,88 @@ record = mapo.matic(features='attr set')
 automap = record.matic(features='auto')
 
 
+class DeferredModule(object):
+
+    class marker_cls(str):
+
+        def __getattr__(self, key):
+            return self.__class__(self + ':' + key)
+
+    class loader_cls(object):
+
+        def load_module(self, name):
+            module = sys.modules[name] = DeferredModule.marker_cls(name)
+            return module
+
+    class dict_cls(dict):
+
+        def __init__(self, *args, **kwds):
+            self.deferred = dict()
+            super(DeferredModule.dict_cls, self).__init__(*args, **kwds)
+
+        def import_one(self, key, name, attr=None):
+            # maybe drop fake and reload real module
+            if attr is None and ':' in name:
+                name, attr = name.rsplit(':', 1)
+            resolved = sys.modules[name]
+            if isinstance(resolved, DeferredModule.marker_cls):
+                sys.modules.pop(name)
+                # find a real module
+                resolved = module = __import__(name, fromlist='bleh')
+                if attr:
+                    resolved = getattr(module, attr)
+                # update the namespace
+                self[key] = resolved
+            return resolved
+
+        def import_all(self):
+            # see __missing__ for why this exists
+            #FIXME: needs ordereddict or similar here
+            for key, name in self.deferred.items():
+                attr = None
+                yield self.import_one(key, name, attr)
+
+        def __missing__(self, key):
+            name = self.deferred.pop(key, None)
+            if name:
+                # on-demand loading seems to only work for code executing at
+                # the module level. a function defined therein is properly
+                # bound (fun.__globals__ is self), but DOES NOT trigger
+                # __missing__ during symbol resolution, ultimately failing with
+                # NameError... as such, xacto preloads all imports once a
+                # tool is "selected".
+                msg = 'early import in {0}: {1}'.format(
+                    self['__name__'], name.replace(':', '.'),
+                    )
+                warnings.warn(msg, ImportWarning, stacklevel=2)
+                return self.import_one(key, name)
+
+            try:
+                return super(DeferredModule.dict_cls, self).__missing__(key)
+            except AttributeError:
+                raise KeyError(key)
+
+        def __setitem__(self, key, item):
+            # if it's a placeholder, don't actually save it... the first
+            # request will then fire __missing__ instead
+            if isinstance(item, DeferredModule.marker_cls):
+                self.deferred[key] = item
+                return None
+
+            return super(DeferredModule.dict_cls, self).__setitem__(key, item)
+
+
+    def __new__(cls, *args, **kwds):
+        self = super(DeferredModule, cls).__new__(cls, *args, **kwds)
+        self.__dict__ = self.dict_cls()
+        return self
+
+    def __init__(self, name, doc=None):
+        super(DeferredModule, self).__init__()
+        self.__name__ = name
+        self.__doc__ = doc
+
+
 class Namespace(object):
 
     def __init__(self, ns, hint=None, parent=None, **kwds):
@@ -228,7 +307,7 @@ class Namespace(object):
             if getattr(module, '__doc__', None):
                 docstr = module.__doc__.strip()
         else:
-            module = self._module = types.ModuleType(str(self))
+            module = self._module = DeferredModule(str(self))
             module.__package__ = module.__name__
             module.__file__ = None
             if isinstance(hint, types.CodeType):
@@ -291,7 +370,7 @@ class Namespace(object):
             module = sys.modules[str(loader)] = loader._module
             try:
                 if loader._code not in (None, True, False):
-                    exec loader._code in module.__dict__
+                    eval(loader._code, module.__dict__)
             except:
                 sys.modules.pop(module.__name__)
                 raise
@@ -302,7 +381,7 @@ class Namespace(object):
     def walk(self, skips=tuple(), rskips=tuple()):
         if self not in skips:
             yield self
-        for key, gen1 in sorted(vars(self).iteritems()):
+        for key, gen1 in sorted(vars(self).items()):
             if not (gen1 in rskips or key.startswith('_')):
                 if isinstance(gen1, Namespace):
                     for gen2 in gen1.walk():
@@ -401,6 +480,7 @@ class Xacto(object):
         self._ns = ns
         self._parser = None
         self._subparser = None
+        self._deferred = dict()
 
         nsroot = self._ns.root
         nspath = self._ns.path
@@ -423,6 +503,9 @@ class Xacto(object):
                         pth.normpath(lname),
                         ).strip('.').replace('/', '.')
                 lname = re.sub('(?<=[.])[0-9]', '_\g<0>', lname)
+                if lname in nsroot._cache:
+                    # node already exists!
+                    continue
 
                 node = None
                 dyn = list()
@@ -436,10 +519,6 @@ class Xacto(object):
 
                     dyn.append(sfx)
 
-                if str(node) == lname:
-                    # node already exists!
-                    continue
-
                 for missing in dyn[::-1]:
                     node = node(missing)
 
@@ -448,8 +527,18 @@ class Xacto(object):
     def find_module(self, name, path=None):
         loader = self._ns.find_module(name)
         if not loader:
-            msg = 'invalid %s tool: %s' % (self._ns._ns, name)
-            warnings.warn(msg, ImportWarning, stacklevel=2)
+            deferred = getattr(self, '_deferred', None)
+            if deferred is not None:
+                if deferred.pop(name, None):
+                    # already deferred once, perm drop the cache
+                    del self._deferred
+                    # and fall thru to default importer
+                    return None
+
+                loader = DeferredModule.loader_cls()
+                self._deferred[name] = loader
+
+        # None == try next importer
         return loader
 
     def load_module(self, name):
@@ -466,7 +555,7 @@ class Xacto(object):
             }
         name_re = re.compile('|'.join(name_map), re.IGNORECASE)
         kill_re = re.compile('|'.join(
-            kv[0] for kv in self.trans.iteritems() if kv[0] and kv[1] is None
+            kv[0] for kv in self.trans.items() if kv[0] and kv[1] is None
             ), re.IGNORECASE)
 
         #FIXME: impl ctx manager for sys.meta_path/sys.modules
@@ -509,12 +598,24 @@ class Xacto(object):
                         help=':= { .%s }' % ctx,
                         )
 
-                _rawdoc = getattr(potential, '__doc__', None)
+                potential_call = getattr(potential, '__call__', None)
+                potential_self = getattr(potential_call, '__self__', None)
+                if potential_self is potential:
+                    # a free function, neither bound nor unbound
+                    potential_call = potential
+                    potential_self = False
+                potential_args = None
+                potential_defs = None
+
+                _rawdoc = (
+                    getattr(potential, '__doc__', None) or
+                    getattr(potential_call, '__doc__', None)
+                    )
                 doc = name_re.sub(name_get, str(trans(_rawdoc)))
                 doc = list(doc.partition('\n')[::2])
                 doc[0] = kill_re.sub('', doc[0])
                 doc[0] = re.sub(' {2,}', ' ', doc[0])
-                doc = map(trans, doc)
+                doc = list(map(trans, doc))
                 doc0 = _rawdoc and doc[0] or None
                 doc01 = doc[0] or doc[1]
                 doc10 = doc[1] or doc[0]
@@ -524,39 +625,36 @@ class Xacto(object):
                 par.epilog = doc0
                 par.xacto = self
                 par._actions[0].help = argparse.SUPPRESS
-                if '-' in app:
+                if '-' in app and sub is not self:
                     sub._name_parser_map[app.replace('-', '_')] = par
                 if not key:
                     orphans.add(ns)
                     continue
 
-                key = None
                 spec = None
-                offset = None
-                attrs = [(None, 0)]
-                if isinstance(potential, type):
-                    attrs.append(('__init__', 1))
-                    attrs.append(('__new__', 1))
-                attrs.append(('__call__', 0))
-                for key, offset in attrs:
-                    try:
-                        attr = key and getattr(potential, key)
-                        spec = inspect.getargspec(attr or potential)
-                    except OSError:
-                        continue
-                    except TypeError:
-                        break
-                    else:
-                        break
+                try:
+                    spec = inspect.getargspec(potential_call)
+                except TypeError:
+                    pass
+                except OSError:
+                    pass
                 if not spec:
                     continue
 
                 nil = object()
                 if spec.defaults is None:
                     spec = spec._replace(defaults=[])
-                defaults = [nil]*(len(spec.args)-len(spec.defaults))
-                defaults.extend(spec.defaults)
-                for arg, default in zip(spec.args, defaults):
+                potential_args = spec.args[potential_self is not False:]
+                potential_defs = spec.defaults[:]
+                potential_diff = len(potential_defs) - len(potential_args)
+                if potential_diff > 0:
+                    potential_defs = potential_defs[potential_diff:]
+                defaults = [nil]*(len(potential_args)-len(potential_defs))
+                defaults.extend(potential_defs)
+                if spec.varargs:
+                    potential_args.append(spec.varargs)
+                    defaults.append(nil)
+                for arg, default in zip(potential_args, defaults):
                     kwds = {
                         'help': trans(None),
                         'dest': arg,
@@ -576,23 +674,127 @@ class Xacto(object):
                         elif default is not None:
                             kwds['type'] = default.__class__
                             kwds['metavar'] = kwds['type'].__name__.upper()
-                    if hasattr(default, '_argspec'):
-                        arginst = default(default, par, potential, arg, kwds)
-                    else:
-                        pfx = '-' * min(len(arg), 2)
-                        arginst = par.add_argument(pfx + arg, **kwds)
-                    setattr(potential, arg, arginst)
-                if spec.varargs:
-                    arginst = par.add_argument(
-                        'xacto_vary',
-                        metavar=spec.varargs,
-                        nargs='*',
-                        )
+
+                    #FIXME: factor this out, almost copypasta with above
+                    proto_attr = getattr(potential, arg, None)
+                    proto_call = getattr(proto_attr, '__call__', None)
+                    proto_self = getattr(proto_call, '__self__', None)
+                    if proto_self is proto_attr:
+                        proto_call = proto_attr
+                        #FIXME: this feels too brittle
+                        if getattr(proto_attr, 'im_class', None) is potential:
+                            proto_self = potential_self
+                    proto_spec = None
+                    proto_args = None
+                    proto_defs = None
+                    proto_doc = None
+
+                    if proto_call:
+                        proto_spec = inspect.getargspec(proto_call)
+                        if proto_spec.defaults is None:
+                            proto_spec = proto_spec._replace(defaults=[])
+                        proto_args = proto_spec.args[proto_self is not False:]
+                        proto_defs = proto_spec.defaults[:]
+                        proto_diff = len(proto_defs) - len(proto_args)
+                        if proto_diff > 0:
+                            proto_defs = proto_defs[proto_diff:]
+                        if proto_args and not proto_spec.varargs:
+                            kwds['nargs'] = len(proto_args)
+                            if kwds['nargs'] == 1:
+                                if proto_defs:
+                                    kwds['const'] = proto_defs[0]
+                                    kwds['nargs'] = '?'
+                                else:
+                                    # don't produce a list of one item
+                                    kwds['nargs'] = None
+                        elif not proto_args and proto_spec.varargs:
+                            kwds['nargs'] = '*'
+                        elif proto_args and proto_spec.varargs:
+                            #TODO: this should really be something like {2,3}
+                            # http://bugs.python.org/issue11354
+                            kwds['nargs'] = '+'
+                        if kwds.get('nargs') is not None:
+                            proto_type_args = list(proto_args)
+                            if proto_spec.varargs:
+                                # special acccessor for *args
+                                slicer = slice(len(proto_args), None)
+                                getter = property(operator.itemgetter(slicer))
+                                proto_type_args.append(proto_spec.varargs)
+                            proto_type = collections.namedtuple(
+                                proto_call.__name__,
+                                proto_type_args,
+                                )
+                            if proto_spec.varargs:
+                                # special acccessor for *args
+                                setattr(proto_type, proto_spec.varargs, getter)
+                            def proto_type_new(cls, *args):
+                                diff = len(proto_args) - len(args)
+                                if diff > 0:
+                                    #FIXME: do this differently...?
+                                    args += (None,) * diff
+                                return tuple.__new__(cls, args)
+                            def proto_type_repr(self, *args):
+                                qualname = '.'.join([
+                                    self.__class__.__module__, name,
+                                    self.__class__.__name__, 'type',
+                                    ])
+                                fields = ', '.join(
+                                    '{0}={1!r}'.format(f, getattr(self, f))
+                                    for f in self._fields
+                                    )
+                                return '{0}({1})'.format(qualname, fields)
+                            proto_type.__module__ = str(ns)
+                            # originals fail with variable-length input
+                            # and repr provides more accurate information
+                            proto_type.__new__ = staticmethod(proto_type_new)
+                            proto_type.__repr__ = proto_type_repr
+                            #FIXME: this REALLY needs to be object based!
+                            # make the generated type available
+                            proto_fun = getattr(
+                                proto_call, '__func__', proto_call,
+                                )
+                            setattr(proto_fun, 'type', proto_type)
+                        proto_doc = trans(proto_call.__doc__)
+                        if proto_doc:
+                            kwds['help'] = proto_doc
+                        if proto_spec.varargs:
+                            #TODO: add support for multiple metavars
+                            kwds['metavar'] = proto_spec.varargs.upper()
+                        elif len(proto_args) == 1:
+                            #TODO: add support for multiple metavars
+                            kwds['metavar'] = proto_args[0].upper()
+                        #elif proto_args:
+                        #    #TODO: add support for multiple metavars
+                        #    kwds['metavar'] = proto_args[0].upper()
+                        #FIXME: detect/handle better... incompatible signature
+                        # eg. cls.__call__(arg=True, ...) and cls.arg(*list)
+                        kwds.pop('action', None)
+
+                    pfx = '-' * min(len(arg), 2)
+                    if arg == spec.varargs:
+                        #TODO: *args (positionals) will ALWAYS be a tuple to
+                        # the final __call__(...); maybe we should chunk them
+                        # into appropriate groups and call proto_type(*group)?
+                        # else *args will ALWAYS be a tuple of length 1 with a
+                        # single proto_type(...) object?
+                        kwds['nargs'] = kwds.pop('nargs', None) or '*'
+                        kwds.pop('required', None)
+                        kwds.pop('metavar', None)
+                        kwds.pop('dest', None)
+                        kwds.pop('type', None)
+                        pfx = ''
+                    arginst = par.add_argument(pfx + arg, **kwds)
+
+                # False : function
+                #  None : method (unbound)
+                #     * : method (bound)
+                if potential_self is None:
+                    potential = potential()
+
                 par.set_defaults(
                     xacto_ns=ns,
                     xacto_call=potential,
-                    xacto_args=spec.args,
-                    xacto_vary=list(),
+                    xacto_args=potential_args,
                     )
         #FIXME: this only works because we [currently] preload all modules
         sys.meta_path.remove(self)
@@ -617,9 +819,21 @@ class Xacto(object):
 
         if module is None:
             module = __name__
-        if isinstance(module, basestring):
+        # TODO: should probably depend on six
+        if isinstance(module, (type(u''), type(''))):
             module = __import__(module, fromlist='*')
 
+        # FIXME: fixup __file__??? this initializer is too gnarly :(
+        if not hasattr(module, '__file__') and hasattr(rel, '__file__'):
+            module.__file__ = rel.__file__
+
+        if name is None:
+            try:
+                # set by zippy; records the resolved export key/entry
+                name = sys.export.key
+            except AttributeError:
+                pass
+
         if None in (path, name):
             file = (
                 getattr(rel, '__file__', '') or
@@ -627,6 +841,8 @@ class Xacto(object):
                 )
             if module.__name__ == '__main__':
                 file = file or sys.argv[0]
+            if file == '-c':
+                file = os.path.basename(sys.executable)
             if file.endswith(('/__init__.py', '/__main__.py')):
                 file = file[:-12]
             if file.endswith(('/__init__.pyc', '/__main__.pyc')):
@@ -641,6 +857,20 @@ class Xacto(object):
         if not name:
             name = pth.basename(file_noext)
 
+        if rel is None:
+            try:
+                rel = __import__(name + '.tools', fromlist='*')
+            except ImportError:
+                pass
+            else:
+                # found {name}.tools
+                return cls.from_module(
+                    module=module, name=rel.__name__, rel=rel,
+                    )
+
+        if not path and rel is not None:
+            path = file_noext
+
         if not path:
             head = pth.dirname(file_noext)
             vfile = pth.join(head, name)
@@ -720,15 +950,38 @@ class Xacto(object):
         if not ns:
             return self._parser.error('no exports found')
 
-        imports = getattr(ns._module, '_imports', None)
-        if imports:
-            ns._module.__dict__.update(imports())
-        call = kwds.pop('xacto_call')
-        args = list(
-            kwds.pop(x)
-            for x in kwds.pop('xacto_args')
-            ) + kwds.pop('xacto_vary')
-        return call(*args, **kwds)
+        import_all = getattr(ns._module.__dict__, 'import_all', None)
+        if import_all:
+            # see DeferredModule.dict_cls.__missing__
+            tuple(import_all())
+
+        args = list()
+        xacto_call = kwds.pop('xacto_call')
+        xacto_args = kwds.pop('xacto_args')
+        for arg_name in xacto_args:
+            arg = kwds.pop(arg_name)
+            arg_new = None
+            arg_prep = getattr(xacto_call, arg_name, None)
+            arg_type = getattr(arg_prep, 'type', None)
+            if not hasattr(arg_prep, '__call__'):
+                arg_prep = None
+            if not hasattr(arg_type, '__call__'):
+                arg_type = None
+            if arg_type:
+                #TODO: only list objects get unpacked? eliminate this?
+                # call this twice because arg_type might fill in any gaps
+                arg = arg_type(*arg)
+            if arg_prep:
+                #TODO: only list objects get unpacked? eliminate this?
+                arg_new = arg_prep(*arg) if arg_type else arg_prep(arg)
+            if arg_new is not None:
+                arg = arg_new
+                if arg_type:
+                    arg = arg_type(*arg)
+            args.append(arg)
+
+        rv = xacto_call(*args, **kwds)
+        return rv
 
 
 class XactoParser(argparse.ArgumentParser):
